"""
    
"""
import StringIO
from DateTime import DateTime

#zope imports
from Acquisition import Implicit
from Globals import Persistent
from AccessControl import ClassSecurityInfo, Owned

# product import
from Products.ploneboard_anyxmlimport.config import HASLIBXML2, PROJECTNAME, \
     VERSION, HASPLONEBOARD, DEFAULTXSLTFILENAME, DEFAULTXMLFILENAME
from Products.ploneboard_anyxmlimport.utils import makeProductPathAbsolute, log

if HASLIBXML2:
    import libxml2
    import libxslt

if HASPLONEBOARD:
    import Products.Ploneboard.catalog




class ploneboardAnyXmlImportTool(Implicit, Owned.Owned):

    security = ClassSecurityInfo()
    security.declareObjectPublic()

    __instance = None #singleton implementation


    class __impl:

        _islibxml2libxsltinstalled = HASLIBXML2
        _xmldoc = None
        _xsltdoc = None
        _xmlintermediary = None
        _output = StringIO.StringIO()
        _displaymodus = "threaded_conversation_view" #"conversation_view" "threaded_conversation_view"
        _xsltfile = DEFAULTXSLTFILENAME
        _includeconvdescriptions = False


        def __init__(self, container):
            self.container = container
            _output = StringIO.StringIO()


        def addOutput(self, anything):
            """ add text to the logger of the migration
            """
            log("addoutput: %s" %anything)
            self._output.write(anything)
            #self.__p_changed = 1


        def getOutput(self):
            """
	       return the log of the migration
            """
            return self._output.getvalue()

    
        def version(self):
            """ return the version """
            return  "%s, version %s" %(PROJECTNAME, VERSION)


        def islibxml2libxsltinstalled(self):
            """ return true if libxml2 and libxslt are correctly installed """
            return self._islibxml2libxsltinstalled


        def setViewAsThreaded(self, booloption):
            """ if this is set to true, then the conversations that are created 
	        will have the "threaded" layout (displayed as indented messages 
		in the conversation tree.
            """
            if booloption:
                self._displaymodus = "threaded_conversation_view"
            else:
                self._displaymodus = "conversation_view"


        def getIsViewAsThreaded(self):
            """ if this returns true, then the conversations that are created will
	        have the "threaded" layout (displayed as indented messages in the
		conversation tree.
            """
            return (self._displaymodus == "threaded_conversation_view")


        def setXsltFilename(self, filename):
            """ the name of the xsltfile that will be applied to transform the
                xml into the xml format that is used internally to create the 
                Ploneboard objects
                The file has to reside in the /xslt directory of this product.
            """
            self._xsltfile = filename


        def getXsltFilename(self):
            """ returns the name of the xsltfile that will be applied to transform 
                the xml into the xml format that is used internally to create  
                the Ploneboard objects
            """
            return self._xsltfile


        def _parseFile(self, filename):
            """
                wrapper for libxml2.parseFile
            """
            try:
                return libxml2.parseFile(filename)
            except Exception, e:
                raise
                return None


        def _parseXml(self, filename, isxslt, subdir):
            """
	        A wrapper around libxml2.parseFile;
            """
            if self._islibxml2libxsltinstalled:
                absfilepath = makeProductPathAbsolute(filename, subdir)
                self.addOutput("<br />parsing file &lt;%s&gt; ...." %absfilepath)
                obj = libxml2.parseFile(absfilepath)
                if isxslt:
                    self._xsltdoc = obj
                else:
                    self._xmldoc = obj
                if obj != None: 
                    self.addOutput("...Done")
                    return True
                else: 
                    return False



        def parse_stylesheet(self, xsltfilename):
            """
                a wrapper for libxslt.parseStylesheetDoc
            """
            if self._parseXml(xsltfilename, True, "xslt"):
                resultingstyle = libxslt.parseStylesheetDoc(self._xsltdoc)
                return resultingstyle
            else: return None


        def freeDocs(self):
            pass
            """ this makes the cpu spin.. dunno why
            try:
                if self._islibxml2libxsltinstalled:
                    if self._xmldoc != None: self._xmldoc.freeDoc()
                    if self._xsltdoc != None: self._xsltdoc.freeDoc()
                    return True
            except:
                raise
            """

        def transformXml(self, xsltfile, xmlfile):
            """
	       a wrapper around some libxml2 / libxslt functions, to apply the
	       provided xslt stylesheet to the provided xml data in order to
	       transform the xml data into the format that is natively used 
	       (see in the example directory for an example)
            """
            try:
                if self._parseXml(xmlfile, False, "inputxml"):
                    style = self.parse_stylesheet(self._xsltfile)
                    self.addOutput("<br />transforming xml into easier-to-handle xml....")
                    self._xmlintermediary = style.applyStylesheet(self._xmldoc, None)
                    #style.saveResultToFilename( makeProductPathAbsolute("intermediatexml.txt", None), self._xmlintermediary, 0 )
                    style.freeStylesheet()
                    self.addOutput("Done")
                    return True
            except Exception, e:
                self.addOutput("<br />transformXml: error %s" %e)
                return False

        def _getormakePloneBoardInstance(self, instanceid, instanceprops):
            """
                make a new ploneboard instance or return an existing one with
                the same id as stated in instanceprops.get('id')
                instanceid = a Plone Site instance, or path within
            """
            try:
                #log("parent: %s; id: %s; props: %s" %(self.container, instanceid, instanceprops) )
                if self.container.get(instanceid, None) != None:
                    return self.container[instanceid]
                else:
                    portal_types = self.container.portal_types
                    portal_types.constructContent('Ploneboard', self.container, instanceid, None)
                    newplbinstance = self.container[instanceid]
                    newplbinstance.title = instanceprops.get('title')
                    return newplbinstance
            except Exception, e:
                self.addOutput("<br />getormakePloneBoardInstance exception: %s" %e)
                return None

        def _getormakePloneboardForum(self, plboardinstance, forumid, forumprops):
            """
                create or return a PloneboardForum with id as specified in param
		forumid.
            """
            try:
                if plboardinstance.get(forumid, None) != None:
                    return plboardinstance[forumid]
                else:
                    portal_types = plboardinstance.portal_types
                    portal_types.constructContent('PloneboardForum', plboardinstance, forumid, None)
                    foruminstance = plboardinstance[forumid]
                    foruminstance.title = forumprops.get('name')
                    foruminstance._setPortalTypeName('PloneboardForum')
                    foruminstance.created = DateTime(forumprops.get('crdate'))
                    return foruminstance
            except Exception, e:
                self.addOutput('<br />getormakePloneboardForum exception: %s' %e)
                return None

        def _getormakePloneboardConversation(self, foruminstance, topicnode, topicprops):
            """
                Start a new Ploneboard conversation (always together with a first 
                posting (Ploneboard Comment), or return an existing conversation
                with the same id.
                param foruminstance is a PloneboardForum object
            """
            try:
                #log("getormakePloneboardConversation %s; props: %s" %(topicnode, topicprops) )
                if foruminstance.get(topicprops.get('id'), None) != None:
                    return foruminstance[topicprops.get('id')]
                else:
                    #foruminstance.addConversation(topicnode, topicprops.get('description'), 
                    portal_types = foruminstance.portal_types
                    portal_types.constructContent('PloneboardConversation', foruminstance, topicprops.get('id'), None)
                    convinstance = foruminstance[topicprops.get('id')]
                    convinstance.title = topicprops.get('title')
                    if self._includeconvdescriptions:
                        convinstance.description = topicprops.get('description')
                    convinstance.setCreators([topicprops.get('creator')])
                    self.addOutput('<br />adding conversation: %s' %convinstance.title)
                    convinstance.setLayout(self._displaymodus)

                    textnode = self._getInnerNode(topicnode, "text")
                    commentrootinstance = convinstance.addComment(topicprops.get('title'), textnode.content, topicprops.get('creator'))
                    commentrootinstance.creation_date = DateTime(topicprops.get('crdate'))
                    commentrootinstance.setModificationDate( DateTime(topicprops.get('modification_date')) )
                    return commentrootinstance
            except Exception, e:
                self.addOutput('<br />getormakePloneboardConversation exception: %s' %e)
                return None


        def _getormakePloneboardReply(self, replyrootinstance, replynode, replyprops):
            """
                Add a new Ploneboard comment as a reply to an existing message 
                (as depicted by replyrootinstance, which is a Ploneboard Comment 
                too), or return the Ploneboard Comment object with that id.
                The replynode is the xml node that contains all the info for the
		new comment, and replyprops is a dict with all its attributes and
		values.
                This function is called recursively for the innernodes of replynode
                (which are replies on the comment to be created)
            """
            try:
                if replyrootinstance.get(replyprops.get('id'), None) != None:
                    replyinstance = replyrootinstance[replyprops.get('id')]
                else:
                    textnode = self._getInnerNode(replynode, "text")
                    replyinstance = replyrootinstance.addReply(replyprops.get('title'), textnode.content, replyprops.get('creator'))
                    replyinstance.creation_date = DateTime(replyprops.get('crdate'))
                    replyinstance.setModificationDate( DateTime(replyprops.get('modification_date')) )

                # traverse to replies on this node
                replynodes = self._getInnerNode(replynode, "reply_postings")
                replynodechild = replynodes.children
                while replynodechild is not None:
                    replychildprops = self._xml_attributes_dict(replynodechild)
                    self._getormakePloneboardReply(replyinstance, replynodechild, replychildprops)
                    replynodechild = replynodechild.next
            except Exception, e:
                self.addOutput('<br />getormakePloneboardReply exception: %s' %e)
                return None


        def _xml_attributes(self, node):
            """
                returns a list of name/value tuples for all the attributes of 
                a xml element node
            """
            prop = node.get_properties()
            while prop:
                yield prop.get_name(), prop.get_content()
                prop = prop.get_next()

        def _xml_attributes_dict(self, node):
            """
                returns the attributes of an xml element as a dictionary
            """
            props = {}
            for name, value in self._xml_attributes(node):
                props[name] = value
            return props

        def _getForumsRootnode(self):
            """ 
                returns the node with name "forumsroot" (should be first node)
                else return None
            """
            rootnode = self._xmlintermediary.children
            if rootnode.name == "forumsroot":
                return rootnode
            else:
                return None


        def _getInnerNode(self, node, innernodename):
            """
                for a given node, return the innernode with name as specified
                in parameter innernodename
            """
            innernode = node.children
            if innernode is None:
                return None
            else:
                while innernode is not None:
                    if innernode.name == innernodename:
                       return innernode
                    innernode = innernode.next


        def migrate(self):
            """
              migrate() can be called after the xml to be imported has been parsed, 
	      transformed into a native xml structure by the stylesheet.
	      see PloneboardXMLImportController.migrate() for that.
              Here, the ploneboard objects are created from the internal, native
	      xml code.	      
	    """

            try:
                plboardnode = self._getForumsRootnode()
                if plboardnode != None:
                    plboardprops = self._xml_attributes_dict(plboardnode)

                    self.addOutput("<br />going to create (or find an existing) instance %s ...." %plboardprops['name'])
                    pbinstance = self._getormakePloneBoardInstance(plboardprops['id'], plboardprops)
                    if pbinstance == None:
                        self.addOutput("not succeeded")
                    else:
                        self.addOutput("Done!")
                        self.addOutput("<br />creating forums...")
                        forumnode = plboardnode.children
                        while forumnode is not None:
                            if forumnode.name == 'forum':
                                forumprops = self._xml_attributes_dict(forumnode)
                                foruminstance = self._getormakePloneboardForum(pbinstance, forumprops.get('id'), forumprops)
                                conversationnode = forumnode.children.children #first element of "posting" within container element "forumpostings"
                                while conversationnode is not None:
                                    if conversationnode.name == 'topic':
                                        convprops = self._xml_attributes_dict(conversationnode)
                                        convrootcommentinstance = self._getormakePloneboardConversation(foruminstance, conversationnode, convprops)
                                        
                                        # get the "<postings>" node (under <topic><forumpostings>)
                                        repliesnode = self._getInnerNode(conversationnode, "forumpostings")
                                        if repliesnode is not None:
                                            replynode = repliesnode.children
                                            while replynode is not None:
                                                replyprops = self._xml_attributes_dict(replynode)
                                                self._getormakePloneboardReply(convrootcommentinstance, replynode, replyprops)
                                                replynode = replynode.next

                                    conversationnode = conversationnode.next

                            forumnode = forumnode.next

                        #pbinstance.reindexObject()
                        try:
                            self.addOutput("<br />reindexing Ploneboard instance.....")
                            pbinstance.portal_catalog.refreshCatalog(clear=True)
                            self.addOutput("...Done")
                        except Exception, e:
                            self.addOutput("<br />Error..%s" %e)

                    return True
            except Exception, e:
                raise

    #################################

    def __init__(self, container):
        """ Create Singleton instance for this object)"""
        if ploneboardAnyXmlImportTool.__instance == None:
            ploneboardAnyXmlImportTool.__instance = ploneboardAnyXmlImportTool.__impl(container)


    def __getattr__(self, attr):
        """Delegate access to implementation"""
        return getattr(self.__instance, attr)

    def __setattr__(self, attr, value):
        """Delegate access to implementation"""
        return setattr(self.__instance, attr, value)









